package main.java.weka.classifiers.functions;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;

import binAutoEncoder.MatrixPair;
import binAutoEncoder.Pair;
import binAutoEncoder.baUtils;
import binAutoEncoder.biguint;
import binAutoEncoder.Zenum;

import com.google.common.primitives.Doubles;

import weka.attributeSelection.PrincipalComponents;
import weka.classifiers.AbstractClassifier;
import weka.classifiers.functions.LibLINEAR;
import weka.classifiers.functions.LinearRegression;
import weka.core.Capabilities;
import weka.core.Capabilities.Capability;
import weka.core.matrix.Maths;
import weka.core.matrix.Matrix;
import weka.core.matrix.SingularValueDecomposition;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.Utils;

public class ba extends AbstractClassifier implements OptionHandler{
	
	//Decoder
	private Vector<LinearRegression> f = new Vector();
	private LinearRegression currentF = new LinearRegression();
	
	//Encoder
	private Vector<LibLINEAR> h = new Vector();
	private LibLINEAR currentH = new LibLINEAR();
	
	//Parameters
	private double[] mu = {0, 0, 0, 0.01, 0.01, 0.02};
	
	private int L;
	
	private int enumVal;
	
	private double norm_range;
	
	private double[] norm_min;
	
	private int V;
	
	public ba(){
		L = 8; enumVal = 8
				; V = 200; this.setMu("0 0 0 0.01 0.02 0.03");
	}
	private void setClassifierParameters() throws Exception{
		String[] fOptions = {"-S","1", "-C", "-minimal", "-R", "0", "-do-not-check-capabilities"};
		String[] hOptions = {"-S","2", "-E","0.001","-B","1","-Q","-C","100","-W" ,"1"};
		currentF.setOptions(fOptions);
		currentH.setOptions(hOptions);
	}
	public Capabilities getCapabilities(){
		Capabilities result = super.getCapabilities();
		result.disableAll();
		
		//Attributes
		result.enable(Capability.NUMERIC_ATTRIBUTES);
		
		//Classes
		result.enable(Capability.BINARY_CLASS);
		
		return result;
	}
	
	public int getNumBits(){
		return L;
	}
	  
	public void setNumBits(int bitNum) {
		// TODO Auto-generated method stub
		L = bitNum;
	}
	
	public int getEnum(){
		return enumVal;
	}
	  
	public void setEnum(int newEnum) {
		enumVal =  newEnum;
	}
	
	public int getV(){
		return this.V;
	}
	
	public void setV(int newV){
		this.V = newV;
	}
	
	public String getMu(){
		StringBuilder build = new StringBuilder();
		for(int i = 0; i < mu.length; i++){
			build.append(mu[i]);
			build.append(" ");
		}
		return build.toString();
	}
	
	public Vector<LinearRegression> getDecoder(){
		return f;
	}
	
	public Vector<LibLINEAR> getEncoder(){
		return h;
	}
	
	public double[] getNormVector(){
		return norm_min;
	}
	
	public double getNormValue(){
		return norm_range;
	}
	
	public void setMu(String muStr){
		StringTokenizer tok = new StringTokenizer(muStr, " ");
		
		if(mu.length == 0){
			System.out.println("Mu not processed, default will be used");
			
		}
		else{
			mu = null;
			mu = new double[tok.countTokens()];
			for(int i = 0; i < mu.length; i++){
				mu[i] = Double.parseDouble(tok.nextToken());
			}
		}
	}
	
	public Enumeration listOptions() {
	    
	    Vector<Option> tempVector = new Vector(2);
	    tempVector.addElement(new Option("\tL: number of bits in the Hamming space (= number of binary hash functions).\n"
				    + "\t(default 8)",
				    "L", 0, "-L <int>"));
	    tempVector.addElement(new Option("\tmu: list of penalty parameter values seperated by spaces; we run one MAC iteration for each."
	    			+ "\t(default {0, 0, 0, 0.01, 0.01, 0.02}",
					"penalty", 0, "-penalty <doubles delimited by spaces>"));
	    tempVector.addElement(new Option("\tenum: if L <= enum, use enumeration (exact) optimization in the Z-step,"
	    			+ "otherwise use alternating optimization. \t(Default: 0.)",
	    			"enum", 1, "-enum <int>"));
	    tempVector.addElement(new Option("\tSet number of points to create validation set\t(Default: 200)",
    				"V", 1, "-V <int>"));
	    return tempVector.elements();
	}
	
	public void setOptions(String[] options) throws Exception {

	    String tempStr = Utils.getOption('L', options);
	    if (tempStr.length() != 0) {
	      setNumBits(new Integer(tempStr).intValue());
	    }
	    else{
	    	setNumBits(6);
	    }
	    tempStr = Utils.getOption("penalty", options);	    
	    if(tempStr.length() != 0){
	    	setMu(tempStr);
	    }
	    else{
	    	setMu("0 0 0 0.01 0.01 0.02 ");
	    }
	    tempStr = Utils.getOption("enum", options);
	    if(tempStr.length() != 0){
	    	setEnum(new Integer(tempStr).intValue());
	    }
	    else{
	    	setEnum(0);
	    }
	    tempStr = Utils.getOption('V', options);
	    if(tempStr.length() != 0){	    	
	    	setV(new Integer(tempStr).intValue() );
	    }
	    else{
	    	setV(200);
	    }
	    
	}
	
	@Override
	public String[] getOptions() {
		// TODO Auto-generated method stub
		List<String> options = new ArrayList<String>();
		
		options.add("-L");
		options.add(" " + getNumBits());
		
		if(getMu().length() != 0){
			options.add("-penalty");
			options.add(" " + getMu());
		}
		
		options.add("-enum");
		options.add(" " + getEnum());
		
		options.add("-V");
		options.add(" " + getV());
		
		return options.toArray(new String[options.size()]);
	}
	
	private double dist(double[] X, double[] V, int D){

		double result = 0;
		for(int i = 0; i < D; i++){
			result+=Maths.square((X[i]-V[i]));
		}
		return result;

	}
	
	private Boolean compareZ(biguint[] Z, biguint[] oldZ){
		for(int i = 0; i < Z.length; i++){
			if(!Z[i].compare(oldZ[i])){
				return false;
			}
		}
		return true;
	}
	
	private double KNNPrecision(biguint[] hX, biguint[] hV, int K, int[][] gt, int Dx){

		int numTest = hV.length;
		int numTrain = hX.length;

		Pair[][] dist1 = new Pair[numTest][numTrain];

		for(int i = 0; i < numTest; i++){
			for(int j = 0; j < numTrain; j++){
				dist1[i][j] = new Pair(j, baUtils.dist(hX[j], hV[i]));
			}
		}

		for(int i = 0; i < numTest; i++){
			Arrays.sort(dist1[i]);
		}


		//Start loop for K here
		int[][] dist = new int[numTest][K];
		for(int i = 0; i < numTest; i++){
			for(int j = 0; j < K; j++){
				dist[i][j] = dist1[i][j].getIndex();
			}
		}

		double precision = 0;
		int goodPairs = 0;
		for(int i = 0; i < numTest; i++){
			for(int j = 0; j < K; j++){
				for(int l=0; l<K; l++)
				{
					if(gt[i][j] == dist[i][l])
					{
							goodPairs++;
							break;
					}
				}				
			}
		}
		precision = (double)goodPairs/(numTest*K);
		return precision;
	}
	
	private Matrix itqHelper(Matrix mZ, int maxit) throws IOException{
		Matrix R = Matrix.random(mZ.getColumnDimension(), mZ.getColumnDimension());
		SingularValueDecomposition svd = R.svd();
		Matrix U11 = svd.getU(); Matrix S2 = svd.getS(); Matrix V2 = svd.getV(); R = U11.copy();
		for(int i = 0; i < maxit; i++){
			Matrix mZZ = mZ.times(R);
			double[][] aUX = new double[mZZ.getRowDimension()][mZZ.getColumnDimension()];
			for(int j = 0; j < mZZ.getRowDimension(); j++){
				Arrays.fill(aUX[j], -1);
				for(int k = 0; k < mZZ.getColumnDimension(); k++){
					if(mZZ.get(j, k) >= 0){
						aUX[j][k] = 1;
					}
				}
			}
			Matrix UX = new Matrix(aUX);
			Matrix C = UX.transpose().times(mZ);
			SingularValueDecomposition csvd = C.svd();
			Matrix UB = csvd.getU();
			Matrix UA = csvd.getV();
			R = UA.times(UB.transpose());
		}
		return R;
	}
	
	private biguint[] ITQ(Instances data, int L, int rot) throws Exception{
		data.setClassIndex(data.numAttributes()-2);
		data.deleteAttributeAt(data.numAttributes()-1);
		int N = data.numInstances();
		int D = data.numAttributes();
		biguint[] rZ = new biguint[N];
		double[] mean = new double[D];
		//mean(X, 1)
		for(int i = 0; i < N; i++){
			DenseInstance tempInst = (DenseInstance) data.get(i);
			for(int j = 0; j < D; j++){
				mean[j] += tempInst.value(j);
			}
		}
		for(int i = 0; i < D; i++){
			mean[i] /= N;
		}
		for(int i = 0; i < N; i++){
			DenseInstance tempInst = (DenseInstance) data.get(i);
			for(int j = 0; j < D; j++){
				tempInst.setValue(j, tempInst.value(j) - mean[j]);
			}
			data.set(i, tempInst);
		}
		Matrix mX = baUtils.instancesToMatrix(data);
		PrincipalComponents pca= new PrincipalComponents();
		String[] pcaOption = {"-R","1","-A", Integer.toString(L), "-C"};
		pca.setOptions(pcaOption);
		pca.buildEvaluator(data);
		double[][] aZ = new double[N][L];
		Instances pcaInst = pca.transformedData(data);
		for(int i = 0; i < N; i++){
			Instance tempInst = pcaInst.get(i);
			for(int j = 0; j < L; j++){
				aZ[i][j] = tempInst.value(j);
			}
		}
		if(rot == 1){
			Matrix mZ = new Matrix(aZ);
			Matrix R = itqHelper(mZ, 50);
			R = mZ.times(R);
			double[][] aR = R.getArray();
			for(int i = 0; i < N; i++){
				rZ[i] = new biguint(L);
				for(int j = 0; j < L; j++){
					if(aR[i][j] >= 0){
						rZ[i].setBit(j, 1);
					}
					else rZ[i].setBit(j, 0);
				}
			}
		}
		else{
			for(int i = 0; i < N; i++){
				rZ[i] = new biguint(L);
				for(int j = 0; j < L; j++){
					if(aZ[i][j] > 0){
						rZ[i].setBit(j, 1);
					}
					else rZ[i].setBit(j, 0);
				}
			}
		}
		return rZ;
	}
	
	public void buildClassifier(Instances dataW) throws Exception {
		setClassifierParameters();
		Instances dataX = new Instances(dataW);
		int N = dataX.numInstances();
		int D = dataX.numAttributes()-1;
		biguint[] Z = ITQ(new Instances(dataX), L, 1);
		Matrix mX = baUtils.instancesToMatrix(dataX);			
		double[][] aX = mX.getArray();		
		double[] max_dims = new double[N];
		double[] min_dims = new double[N];
		double range_dims = Double.NEGATIVE_INFINITY;
		
		//Normalization of X
		for(int i = 0; i < mX.getRowDimension(); i++)
		{
			max_dims[i] = Doubles.max(aX[i]);
			min_dims[i] = Doubles.min(aX[i]);
			if(range_dims < (max_dims[i] - min_dims[i]))
			{
				range_dims = max_dims[i] - min_dims[i];
			}
		}

		if(range_dims == 0)
		{
			range_dims = Double.NEGATIVE_INFINITY;
		}

		for(int i = 0; i < mX.getRowDimension(); i++)
		{
			for(int j = 0; j < D; j++)
			{
				aX[i][j] -= min_dims[j];
				aX[i][j] = aX[i][j]/range_dims;
			}
		}
		
		this.norm_range = range_dims;
		this.norm_min = min_dims;
		
		dataX.delete();
		double[][] aXX = new double[N][D+1];
		for(int i = 0; i < N; i++)
		{
			for(int j = 0; j < D; j++)
			{
				aXX[i][j]=aX[i][j];
			}					
			aXX[i][D]= 0;
			dataX.add(new DenseInstance(1, aXX[i]));
		}
		
		//Generates V
		Instances dataV = new Instances(dataX, V);
		ArrayList<Integer> vIndex = new ArrayList<Integer>();
		for(int i = 0; i < N; i++){
			vIndex.add(new Integer(i));
		}
		Collections.shuffle(vIndex);
		for(int i = 0; i < V; i++)
		{		
			dataV.add(dataX.get(vIndex.get(i)));
		}				
		Matrix mV = baUtils.instancesToMatrix(dataV);
		vIndex = null;
		
		double[][] aV = mV.getArray();
		int k = (int) Math.floor(.02*N);
		Pair[][] gt1 = new Pair[V][N];
		
		//Get sorted validation set
		for(int i = 0; i < V; i++)
		{
			for(int j = 0; j < N; j++)
			{
				gt1[i][j] = new Pair(j, dist(aX[j], aV[i],D));
			}
		}

		for(int i = 0; i < mV.getRowDimension(); i++){
			Arrays.sort(gt1[i]);
		}

		int[][] gt = new int[V][k];
		for(int i = 0; i < V; i++){
			for(int j = 0; j < k; j++){
				gt[i][j] = gt1[i][j].getIndex();
			}
		}
		gt1 = null;
		
		//Create testing set Z in the proper format
		ArrayList<Attribute> attList= new ArrayList<Attribute>();
		for(Integer i = new Integer(0); i < L+1; i++){
			attList.add(new Attribute(i.toString()));
		}
		Instances dataZ = new Instances("dataZ", attList, N);				
		mX = baUtils.instancesToMatrix(dataX);	
		
		//Loop Here
		int count = 0;
		double P = 0;		
		double oldP = -1;		
		biguint[] hV; biguint[] hX = new biguint[N];
		biguint[] oldhX = new biguint[N]; biguint[] oldZ = new biguint[N]; 	Matrix rZ = new Matrix(N, L); Matrix oldrZ = new Matrix(N, L);
		
		for (int m = 0; m < N; m++) {
			oldZ[m] = new biguint(L);
			oldZ[m] = Z[m];
			oldhX[m]= new biguint(L);
		}
		
		Vector<LinearRegression> oldF = new Vector<LinearRegression>(D-1);
		Vector<LibLINEAR> oldH = new Vector<LibLINEAR>(L);		
		dataX.setClassIndex(dataX.numAttributes()-1);
		dataZ.setClassIndex(dataZ.numAttributes()-1);
		dataV.setClassIndex(dataV.numAttributes()-1);
		

		for(int m = 0; m < mu.length; m++){
			count++;			
			System.out.print("\r Iteration #" + count);						
									
			//F-Step (Regression)
			oldF.removeAllElements(); oldF = (Vector<LinearRegression>) f.clone();
			f.removeAllElements();
			double[][] tempZ = new double[N][L+1];
			dataZ.delete();
			for(int i = 0; i < N; i++){
				for(int j = 0; j < L; j++){
					tempZ[i][j] = (double)Z[i].getBit(j);
				}
				tempZ[i][L] = 0;
				dataZ.add(new DenseInstance(1, tempZ[i]));

			}
			tempZ = null;

			
			for(int i = 0; i < D; i++){
				for(int j = 0; j < N; j++){
					DenseInstance tempInst = (DenseInstance)dataZ.get(j);
					tempInst.setValue(dataZ.numAttributes()-1, aX[j][i]);
					dataZ.set(j, tempInst);
				}
				
				currentF.buildClassifier(dataZ);
				f.add(i, (LinearRegression) AbstractClassifier.makeCopy(currentF));
			}
			
			
			//H-Step (Classification)			
			oldH.removeAllElements(); oldH = (Vector<LibLINEAR>) h.clone();
			h.removeAllElements();
			for(int i = 0; i < L; i++){
				for(int j = 0; j < N; j++){
					DenseInstance tempInst = (DenseInstance)dataX.get(j);
					tempInst.setValue(dataX.numAttributes()-1, Integer.toString(Z[j].getBit(i)));
					dataX.set(j, tempInst);
				}
				currentH.buildClassifier(dataX);
				h.add(i, (LibLINEAR) AbstractClassifier.makeCopy(currentH));
			}
			
			for(int i = 0; i < N; i++){				
				oldhX[i] = hX[i];
			}
			//Calculate precision
			hX = baUtils.hClassify(dataX, h);
			hV = baUtils.hClassify(dataV, h);
			double[][] fZ = baUtils.fClassify(dataZ, f);
			
			double[][] temphX = new double[N][L];
			for(int i = 0; i < N; i++)
			{
				for(int j = 0; j < L; j++)
				{
					temphX[i][j] = (double)hX[i].getBit(j);
				}
			}

			double [] coefficients = f.get(0).coefficients();
			P = KNNPrecision(hX, hV, k, gt, D);
			
			//Skip Z-step if precision does not increase
			
			if(P < oldP)
			{
				f.removeAllElements();
				h.removeAllElements();
				for(int i = 0; i < N; i++)
				{
					Z[i] = oldZ[i];
					hX[i]=oldhX[i];
				}
				f = (Vector<LinearRegression>) oldF.clone();
				h = (Vector<LibLINEAR>) oldH.clone();
				rZ = oldrZ.copy();
				P = oldP;
								
			}			
			oldP = P; System.out.print(", precision = " + P);
				
			//Z-step
			for(int i = 0; i < N; i++){
				oldZ[i] = Z[i];	
			}
			oldrZ = rZ.copy();
			if (L <= this.enumVal) {
				Z = Zenum.enumerate(mX, hX, Z, mu[m], f);
			} else 
			{
				MatrixPair zPair = Zenum.Zrelaxed(mX.getMatrix(0, N-1, 0, mX.getColumnDimension()-2), f, hX, mu[m], rZ, 1000, .0001, new Matrix(fZ), m);
				Z = zPair.getZ();
				rZ = zPair.getRelaxedZ();
				Z = Zenum.zaltopt(mX, hX, Z, mu[m], f, 1, 100);
			}						
				
			if (compareZ(hX, Z) == true) 
			{				
				break;
			}																						
		}		
		//System.out.println("Final Precision: " + P);
	}
	
	
	public double classifyInstance(Instance instance) throws Exception{
		if(h.size() == 0){
			throw new RuntimeException("Classifier not built!");
		}
		
		int[] result = new int[h.size()];
		for(int i = 0; i < h.size(); i++){
			result[i] = (int) h.get(i).classifyInstance(instance);
			System.out.print(result[i]);
		}
		System.out.println();
		return h.get(0).classifyInstance(instance);
	}
	
	//For external classes implementing ba
	public int[] encodeDouble(Instance instance) throws Exception{
		if(h.size() == 0){
			throw new RuntimeException("Classifier not built!");
		}
		for(int i = 0; i < instance.numAttributes()-1; i++){
			instance.setValue(i, (instance.value(i) - norm_min[i])/norm_range);
		}
		int[] result = new int[h.size()];
		for(int i = 0; i < h.size(); i++){
			result[i] = (int) h.get(i).classifyInstance(instance);
		}
		return result;
	}
	
	//For external classes implementing ba
	public biguint encodeBui(Instance instance) throws Exception{
		if(h.size() == 0){
			throw new RuntimeException("Classifier not built!");
		}
		for(int i = 0; i < instance.numAttributes()-1; i++){
			instance.setValue(i, (instance.value(i) - norm_min[i])/norm_range);
		}
		biguint result = new biguint(h.size());
		for(int i = 0; i < h.size(); i++){
			result.setBit(i, (int)h.get(i).classifyInstance(instance));
		}
		return result;
	}
	
	//Returns normalized values, external classes implementing ba
	public double[] decoder(Instance instance) throws Exception{
		if(f.size() == 0){
			throw new RuntimeException("Classifier not built!");
		}
		double[] result = new double[f.size()];
		for(int i = 0; i < f.size(); i++){
			result[i] = f.get(i).classifyInstance(instance);
		}
		return result;
	}
	
	public static void main(String[] args) {
		//String[] argv = {"-t", "/home/ramin/SIFTTrain.arff", "-T", "/home/ramin/SIFTTest.arff","-V" , "300","-penalty","0 0 .1","-enum","3"};
		//runClassifier(new ba(), argv);
		runClassifier(new ba(), args);
	}


}
